#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import ABC, abstractmethod

from mesh.kinds import Event
from mesh.macro import mpi, spi


@spi("mesh")
class Subscriber(ABC):

    @abstractmethod
    @mpi("mesh.queue.subscribe")
    def subscribe(self, event: Event):
        """
        Subscribe the event with {@link com.be.mesh.client.annotate.Bindings} or {@link com.be.mesh.client.annotate.Binding}
        :param event:
        :return:
        """
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#


from mesh.macro.index import index


class Cause:

    @index(0)
    def name(self) -> str:
        return ''

    @index(5)
    def pos(self) -> str:
        return ''

    @index(10)
    def text(self) -> str:
        return ''

    @index(15)
    def buff(self) -> bytes:
        return b''

    @staticmethod
    def of(e: BaseException) -> "Cause":
        return Cause()

    @staticmethod
    def of_cause(code: str, message: str, cause: "Cause") -> BaseException:
        raise BaseException("")



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import inspect
from typing import Type

from mesh.cause import ValidationException
from mesh.macro import spi, mpi, T, Proxy, MethodProxy
from mesh.mpc.reference import ReferenceInvokeHandler


@spi("mpc")
class ServiceProxy(MethodProxy):

    def proxy(self, kind: Type[T]) -> T:
        return ServiceProxy.default_proxy(kind)

    @staticmethod
    def default_proxy(kind: Type[T]) -> T:
        return ServiceProxy.static_proxy(kind, mpi())

    @staticmethod
    def static_proxy(kind: Type[T], metadata: mpi) -> T:
        interfaces = [kind]
        if hasattr(kind, "__args__") and getattr(kind, "__args__") is not None:
            interfaces = kind.__args__

        for interface in interfaces:
            if not inspect.isabstract(interface):
                raise ValidationException(f'{str(interface)} is not abstract class. ')

        if metadata is None:
            metadata = mpi()

        return Proxy(interfaces, ReferenceInvokeHandler(metadata, interfaces))



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import collections

import grpc

import mesh.tool as tool
from mesh.mpc import Mesh
from mesh.prsim import Metadata


class ClientCallDetails(
    collections.namedtuple('_ClientCallDetails',
                           ('method', 'timeout', 'metadata', 'credentials',
                            'wait_for_ready', 'compression')), grpc.ClientCallDetails):
    pass


class GrpcInterceptor(grpc.ServerInterceptor,
                      grpc.StreamStreamClientInterceptor,
                      grpc.StreamUnaryClientInterceptor,
                      grpc.UnaryStreamClientInterceptor,
                      grpc.UnaryUnaryClientInterceptor, ):
    TKeys = [
        Metadata.MESH_INCOMING_PROXY,
        Metadata.MESH_OUTGOING_PROXY,
        Metadata.MESH_SUBSET,
        Metadata.MESH_VERSION,
        Metadata.MESH_TIMESTAMP,
        Metadata.MESH_RUN_MODE,
        # INC
        Metadata.MESH_TECH_PROVIDER_CODE,
        Metadata.MESH_TOKEN,
        Metadata.MESH_TARGET_NODE_ID,
        Metadata.MESH_TARGET_INST_ID,
        Metadata.MESH_SESSION_ID
    ]

    @staticmethod
    def client_context(context):
        # python metadata must be lowercase
        attachments = Mesh.context().get_attachments()
        metadata = []
        Metadata.MESH_URN.append(metadata, Mesh.context().get_urn())
        Metadata.MESH_TRACE_ID.append(metadata, Mesh.context().get_trace_id())
        Metadata.MESH_SPAN_ID.append(metadata, Mesh.context().get_span_id())
        Metadata.MESH_FROM_INST_ID.append(metadata, Mesh.context().get_consumer().inst_id)
        Metadata.MESH_FROM_NODE_ID.append(metadata, Mesh.context().get_consumer().node_id)
        Metadata.MESH_INCOMING_HOST.append(metadata, f"{tool.get_mesh_name()}@{str(tool.get_mesh_runtime())}")
        Metadata.MESH_OUTGOING_HOST.append(metadata, attachments.get(Metadata.MESH_INCOMING_HOST.key(), ''))

        for mk in GrpcInterceptor.TKeys:
            mk.append(metadata, attachments.get(mk.key(), ''))

        if context.metadata is not None:
            for x in context.metadata:
                metadata.append(x)

        wait_for_ready = context.wait_for_ready if hasattr(context, 'wait_for_ready') else None
        compression = context.compression if hasattr(context, 'compression') else None
        credentials = context.credentials if hasattr(context, 'credentials') else None
        return ClientCallDetails(context.method, context.timeout, metadata, credentials, wait_for_ready, compression)

    def intercept_stream_stream(self, continuation, client_call_details, request_iterator):
        return continuation(self.client_context(client_call_details), request_iterator)

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        return continuation(self.client_context(client_call_details), request_iterator)

    def intercept_unary_stream(self, continuation, client_call_details, request):
        return continuation(self.client_context(client_call_details), request)

    def intercept_unary_unary(self, continuation, client_call_details, request):
        return continuation(self.client_context(client_call_details), request)

    def intercept_service(self, continuation, handler_call_details):
        self.server_context(handler_call_details)
        return continuation(handler_call_details)

    @staticmethod
    def server_context(handler_call_details):
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import mesh.boost as boost
import mesh.cause as cause
import mesh.codec as codec
import mesh.context as context
import mesh.environ as environ
import mesh.grpx as grpx
import mesh.http as http
import mesh.ioc as ioc
import mesh.kinds as kinds
import mesh.log as log
import mesh.macro as macro
import mesh.metrics as metrics
import mesh.mpc as mpc
import mesh.prsim as prsim
import mesh.runtime as runtime
import mesh.schema as schema
import mesh.system as system
import mesh.tool as tool


def init():
    cause.init()
    codec.init()
    grpx.init()
    http.init()
    ioc.init()
    kinds.init()
    log.init()
    macro.init()
    metrics.init()
    mpc.init()
    prsim.init()
    runtime.init()
    schema.init()
    system.init()
    tool.init()
    boost.init()
    context.init()
    environ.init()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
import threading
from typing import List, Dict

import mesh.log as log
import mesh.tool as tool
from mesh.kinds import Topic, Timeout, Event
from mesh.macro import spi, ServiceLoader, Binding
from mesh.mpc import ServiceProxy
from mesh.prsim import Scheduler, Subscriber


@spi("python")
class PythonScheduler(Scheduler):

    def __init__(self):
        self.remote = ServiceProxy.default_proxy(Scheduler)
        self.tasks: Dict[str, Task] = {}

    def timeout(self, timeout: Timeout, duration: int) -> str:
        return ''

    def cron(self, cron: str, binding: Topic) -> str:
        return ''

    def period(self, duration: int, binding: Topic) -> str:
        task_id = tool.next_id()
        log.info(f"Next task {task_id} has been submit.")
        task = Task(duration / 1000, self.do_emit(binding))
        task.start()
        self.tasks[task_id] = task
        return task_id

    def dump(self) -> List[str]:
        task_ids = []
        for task_id, _ in self.tasks.items():
            task_ids.append(task_id)
        return task_ids

    def cancel(self, task_id: str) -> bool:
        return self.stop(task_id)

    def stop(self, task_id: str) -> bool:
        task = self.tasks.get(task_id, None)
        if task:
            task.cancel()
            self.tasks.__delitem__(task_id)
        return True

    def emit(self, topic: Topic) -> bool:
        subscribers = ServiceLoader.load(Subscriber).list('')
        for subscriber in subscribers:
            try:
                bindings = Binding.get_binding_if_present(subscriber)
                if not topic.matches(bindings):
                    continue
                subscriber.subscribe(Event.new_instance({}, topic))
            except BaseException as e:
                log.error(f"{e}")

        return True

    def do_emit(self, topic: Topic):
        def x():
            self.emit(topic)

        return x

    def shutdown(self, duration: int) -> bool:
        for _, task in self.tasks.items():
            task.cancel()
        self.tasks.clear()
        return True


class Task(threading.Thread):

    def __init__(self, interval, function, args=None, kwargs=None):
        threading.Thread.__init__(self)
        self.interval = interval
        self.function = function
        self.args = args if args is not None else []
        self.kwargs = kwargs if kwargs is not None else {}
        self.finished = threading.Event()

    def cancel(self):
        """Stop the timer if it hasn't finished yet."""
        self.finished.set()

    def run(self):
        while not self.finished.isSet():
            try:
                self.function(*self.args, **self.kwargs)
                self.finished.wait(self.interval)
            except BaseException as e:
                log.error(f"{e}")



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import argparse
import os
import random
import socket
from contextlib import closing
from enum import Enum
from typing import List

from mesh.environ.cache import Cache

DEFAULT_MESH_PORT = 570


class StatefulServer:
    def __init__(self, available: bool, address: str):
        self.available = available
        self.address = address


class Addrs:

    def __init__(self, servers: str):
        self.servers: List[StatefulServer] = []
        self.available_addrs = []
        if not servers:
            return
        for server in servers.split(","):
            if server and server.__len__() > 0 and not self.available_addrs.__contains__(server):
                available_addr = server if server.__contains__(":") else f"{server}:{DEFAULT_MESH_PORT}"
                self.available_addrs.append(available_addr)
                self.servers.append(StatefulServer(True, available_addr))

    def any(self) -> str:
        addrs = self.available_addrs
        if not addrs or addrs.__len__() < 1:
            return ''
        return addrs[random.randint(0, addrs.__len__() - 1)]

    def many(self) -> List[str]:
        addrs = self.available_addrs
        if not addrs or addrs.__len__() < 1:
            return []
        return addrs.copy()

    def all_servers(self) -> List[StatefulServer]:
        return self.servers

    def available(self, addr: str, available: bool):
        for server in self.servers:
            if addr == server.address and server.available is not available:
                server.available = available
                available_addrs = []
                for serv in self.servers:
                    if serv.available:
                        available_addrs.append(serv.address)
                if available_addrs.__len__() > 0:
                    self.available_addrs = available_addrs
                else:
                    available_addrs.append(self.servers[random.randint(0, self.servers.__len__() - 1)].address)
                    self.available_addrs = available_addrs
                return


class URI:
    def __init__(self, hostname: str, port: int):
        self.hostname = hostname
        self.port = port

    def __str__(self) -> str:
        return f"{self.hostname}:{self.port}"

    @staticmethod
    def parse(addr: str) -> "URI":
        if not addr.__contains__(":"):
            return URI(addr, 0)
        pair = addr.split(":")
        return URI(pair[0], int(pair[1]) if pair[1].isdigit() else 0)


class Mode(Enum):
    Disable = 1
    Failfast = 2
    Nolog = 4
    JsonLogFormat = 8
    RCache = 16
    PHeader = 32
    Metrics = 64
    RLog = 128
    MGrpc = 256
    PermitCirculate = 512
    NoStdColor = 1024

    def __init__(self, v: int) -> None:
        self.v = v

    def enable(self) -> bool:
        return self.match(System.mode())

    def match(self, mode: int) -> bool:
        return (self.v & mode) == self.v


class Environ:
    __mesh_address_keys = ['mesh.address', 'mesh_address', 'mesh-address', 'MESH_ADDRESS']
    __mesh_runtime_keys = ['mesh.runtime', 'mesh_runtime', 'mesh-runtime', 'MESH_RUNTIME']
    __mesh_mode_keys = ['mesh.mode', 'mesh_mode', 'mesh-mode', 'MESH_MODE']
    __mesh_name_keys = ['mesh.name', 'mesh_name', 'mesh-name', 'MESH_NAME']
    __mesh_direct_keys = ['mesh.direct', 'mesh_direct', 'mesh-direct', 'MESH_DIRECT']
    __mesh_subset_keys = ['mesh.subset', 'mesh_subset', 'mesh-subset', 'MESH_SUBSET']
    __mesh_min_channels = ['mesh.grpc.channel.min', 'mesh_grpc_channel_min', 'MESH_GRPC_CHANNEL_MIN']
    __mesh_max_channels = ['mesh.grpc.channel.max', 'mesh_grpc_channel_max', 'MESH_GRPC_CHANNEL_MAX']

    @staticmethod
    def get_default_mesh_port() -> int:
        return DEFAULT_MESH_PORT

    def get_property(self, dft: str, keys: List[str]) -> str:
        if not keys:
            return dft
        for key in keys:
            if '' != self.get_arguments.get(key, ''):
                return self.get_arguments[key]
            if os.getenv(key):
                return os.getenv(key)
        return dft

    @Cache
    def get_arguments(self):
        parser = argparse.ArgumentParser()
        parser.add_argument("--mesh.address", type=str, default='', required=False)
        parser.add_argument("--mesh.runtime", type=str, default='', required=False)
        parser.add_argument("--mesh.enable", type=str, default='', required=False)
        parser.add_argument("--mesh.name", type=str, default='', required=False)
        parser.add_argument("--mesh.direct", type=str, default='', required=False)
        parser.add_argument("--mesh.grpc.channel.min", type=int, default=20, required=False)
        parser.add_argument("--mesh.grpc.channel.max", type=int, default=200, required=False)
        args, _ = parser.parse_known_args()
        return vars(args)

    @Cache
    def get_mesh_address(self) -> Addrs:
        return Addrs(self.get_property('127.0.0.1', self.__mesh_address_keys))

    @Cache
    def get_mesh_runtime(self) -> URI:
        mr = self.get_property('', self.__mesh_runtime_keys)
        if mr is not None and '' != mr:
            return URI.parse(mr)
        return URI.parse(f'{self.get_ip()}:{self.get_available_port()}')

    @Cache
    def get_mesh_name(self) -> str:
        return self.get_property('unknown', self.__mesh_name_keys)

    @Cache
    def get_mesh_mode(self) -> str:
        return self.get_property('0', self.__mesh_mode_keys)

    @Cache
    def get_mesh_direct(self) -> str:
        return self.get_property('', self.__mesh_direct_keys)

    @Cache
    def get_mesh_subset(self) -> str:
        return self.get_property('', self.__mesh_subset_keys)

    @Cache
    def get_max_channels(self) -> int:
        return int(self.get_property('200', self.__mesh_max_channels))

    @Cache
    def get_min_channels(self) -> int:
        return int(self.get_property(os.cpu_count() * 2, self.__mesh_min_channels))

    @Cache
    def get_hostname(self) -> str:
        return socket.gethostname()

    @Cache
    def get_ip(self) -> str:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.settimeout(0)
        try:
            # doesn't even have to be reachable
            s.connect(('10.255.255.255', 1))
            ip = s.getsockname()[0]
        except Exception as e:
            print(e)
            ip = '127.0.0.1'
        finally:
            s.close()
        return ip

    @Cache
    def get_ip_hex(self) -> str:
        return '{:02X}{:02X}{:02X}{:02X}'.format(*map(int, self.get_ip().split('.')))

    @Cache
    def get_available_port(self) -> str:
        with closing(socket.socket()) as s:
            s.bind(('', 0))
            return s.getsockname()[1]


__environ__ = Environ()
__mode__ = int(__environ__.get_mesh_mode())


class System:

    @staticmethod
    def mode() -> int:
        return __mode__

    @staticmethod
    def environ() -> Environ:
        return __environ__



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from mesh.macro import index, serializable


@serializable
class Captcha:

    @index(0)
    def mno(self) -> str:
        return ""

    @index(5)
    def kind(self) -> str:
        return ""

    @index(10)
    def mime(self) -> bytes:
        return b''

    @index(15)
    def text(self) -> str:
        return ""



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.macro import serializable


@serializable
class Timeout:
    pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.macro import spi
from mesh.mpc import ServiceProxy
from mesh.prsim import Cluster


@spi("mesh")
class MeshCluster(Cluster):

    def __init__(self):
        self.proxy = ServiceProxy.default_proxy(Cluster)

    def election(self, buff: bytes) -> bytes:
        return self.proxy.election(buff)

    def is_leader(self) -> bool:
        return self.proxy.is_leader()



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from mesh.macro import ServiceLoader
from mesh.runtime.context import MeshContext

from mesh.ioc import PaaSProcessor, EnvsProcessor


class Container:

    def __init__(self):
        self.context = MeshContext()

    def start(self):
        self.scan()

    def stop(self):
        pass

    def wait(self):
        pass

    def scan(self):
        for processor in ServiceLoader.load(EnvsProcessor).list():
            self.context.register_processor(processor)
        for processor in ServiceLoader.load(PaaSProcessor).list():
            self.context.register_processor(processor)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import threading
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, Generic, Deque, Optional, List

import mesh.log as log
from mesh.kinds.location import Location, Principal
from mesh.macro import T


class RunMode(Enum):
    ROUTINE = 1
    """
    正常模式
    """
    PERFORM = 2
    """
    评测模式
    """
    DEFENSE = 4
    """
    高防模式
    """
    DEBUG = 8
    """
    调试模式
    """
    LOAD_TEST = 16
    """
    压测模式
    """
    MOCK = 32
    """
    Mock模式
    """

    def is_debug(self) -> bool:
        return RunMode.matches(RunMode.DEBUG, self.value)

    def is_load_test(self) -> bool:
        return RunMode.matches(RunMode.LOAD_TEST, self.value)

    def is_routine(self) -> bool:
        return RunMode.matches(RunMode.ROUTINE, self.value)

    def is_perform(self) -> bool:
        return RunMode.matches(RunMode.PERFORM, self.value)

    def is_defense(self) -> bool:
        return RunMode.matches(RunMode.DEFENSE, self.value)

    def is_mock(self) -> bool:
        return RunMode.matches(RunMode.MOCK, self.value)

    @staticmethod
    def matches(mode: Any, code: int):
        return isinstance(mode, RunMode) and (mode.value & code) is code

    @staticmethod
    def get_by_code(code: int) -> "RunMode":
        for e in RunMode:
            if (e.value & code) is e.value:
                return e
        return RunMode.ROUTINE


class Key(Generic[T], ABC):
    """
    Context key.
    """

    def __init__(self, name: str):
        self.name = name

    @abstractmethod
    def get_if_absent(self) -> T:
        """ Get the attribute if present by key """
        pass

    @abstractmethod
    def map(self, fn) -> Optional[Any]:
        """ Map new value """
        pass

    @abstractmethod
    def if_present(self, fn):
        """ Consume attribute if present """
        pass

    @abstractmethod
    def or_else(self, v: T) -> T:
        """ Get attribute default if optional """
        pass

    @abstractmethod
    def is_present(self) -> bool:
        """ Check the attribute is present """
        pass


class Metadata(Enum):
    """
    https://www.rfc-editor.org/rfc/rfc7540#section-8.1.2
    """

    MESH_SPAN_ID = "mesh-span-id"
    MESH_TIMESTAMP = "mesh-timestamp"
    MESH_RUN_MODE = "mesh-run-mode"
    MESH_CONSUMER = "mesh-consumer"
    MESH_PROVIDER = "mesh-provider"
    MESH_URN = "mesh-urn"
    #
    MESH_INCOMING_HOST = "mesh-incoming-host"
    MESH_OUTGOING_HOST = "mesh-outgoing-host"
    MESH_INCOMING_PROXY = "mesh-incoming-proxy"
    MESH_OUTGOING_PROXY = "mesh-outgoing-proxy"
    MESH_SUBSET = "mesh-subset"
    # PTP
    MESH_VERSION = "mesh-version"
    MESH_TECH_PROVIDER_CODE = "mesh-tech-provider-code"
    MESH_TRACE_ID = "mesh-trace-id"
    MESH_TOKEN = "mesh-token"
    MESH_FROM_NODE_ID = "mesh-from-node-id"
    MESH_FROM_INST_ID = "mesh-from-inst-id"
    MESH_TARGET_NODE_ID = "mesh-target-node-id"
    MESH_TARGET_INST_ID = "mesh-target-inst-id"
    MESH_SESSION_ID = "mesh-session-id"

    def key(self) -> str:
        return self.value

    def set(self, attachments: Dict[str, str], v: str):
        if attachments and '' != v and v != attachments.get(self.key(), ''):
            attachments[self.key()] = v

    def get(self, attachments: Dict[str, str]) -> str:
        return attachments.get(self.key(), '')

    def append(self, attachments: List[Any], v: str):
        if attachments is not None and '' != v:
            attachments.append((self.key(), v))


class Context(ABC):
    """
    MPC invoke context.
    """

    @abstractmethod
    def get_trace_id(self) -> str:
        """
        Get the request trace id.
        :return: The request trace id.
        """
        pass

    @abstractmethod
    def get_span_id(self) -> str:
        """
        Get the request span id.
        :return: The request span id.
        """
        pass

    @abstractmethod
    def get_timestamp(self) -> int:
        """
        Get the request create time.
        :return: The request create time.
        """
        pass

    @abstractmethod
    def get_run_mode(self) -> RunMode:
        """
        Get the request run mode. RunMode
        :return: The request run mode.
        """
        pass

    @abstractmethod
    def get_urn(self) -> str:
        """
        Mesh resource uniform name. Like: create.tenant.omega.json.http2.lx000001.mpi.trustbe.net
        :return: Uniform name.
        """
        pass

    @abstractmethod
    def get_consumer(self) -> Location:
        """
        Get the consumer network principal.
        :return: Consumer network principal.
        """
        pass

    @abstractmethod
    def get_provider(self) -> Location:
        """
        Get the provider network principal.
        :return: Provider network principal.
        """
        pass

    @abstractmethod
    def get_attachments(self) -> Dict[str, str]:
        """
        Dispatch attachments.
        :return: Dispatch attachments.
        """
        pass

    @abstractmethod
    def get_principals(self) -> "Queue[Principal]":
        """
        Get the mpc broadcast network principals.
        :return: Broadcast principals.
        """
        pass

    @abstractmethod
    def get_attributes(self) -> Dict[str, Any]:
        """
        Get the context attributes. The attributes don't be serialized.
        :return: attributes
        """
        pass

    @abstractmethod
    def get_attribute(self, key: Key[T]) -> T:
        """
        Like getAttachments, but attribute wont be transfer in invoke chain.
        :param key: attribute key
        :return: attribute value
        """
        pass

    @abstractmethod
    def set_attribute(self, key: Key[T], value: T) -> None:
        """
        Like putAttachments, but attribute won't be transfer in invoke chain.
        :param key: attribute key
        :param value: attribute
        :return: None
        """
        pass

    @abstractmethod
    def rewrite_urn(self, urn: str) -> None:
        """
        Rewrite the urn.
        :param urn: urn
        :return:
        """
        pass

    @abstractmethod
    def rewrite_context(self, another: "Context") -> None:
        """
        Rewrite the context by another context.
        :param another: another context
        :return:
        """
        pass

    @abstractmethod
    def resume(self) -> "Context":
        """
        Open a new context.
        :return:
        """
        pass

    @abstractmethod
    def encode(self) -> Dict[str, str]:
        """
        Encode context to string.
        :return:
        """
        pass

    @abstractmethod
    def decode(self, attachments: Dict[str, str]) -> None:
        """
        Decode context from string.
        :return:
        """
        pass


class Queue(Deque, Generic[T]):
    def pop(self) -> T:
        if self.__len__() < 1:
            log.warn(f"{threading.currentThread().ident}({threading.currentThread().name}) pop empty queue.")
            return None
        return super().pop()

    def peek(self) -> T:
        if self.__len__() < 1:
            return None
        return self[-1]



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from typing import List, Dict, Any

import mesh.tool as tool
from mesh.cause import ValidationException
from mesh.context import Mesh
from mesh.kinds import Principal
from mesh.macro import T, spi, Proxy, InvocationHandler
from mesh.prsim import Routable


class MeshRouter(InvocationHandler):

    def __init__(self, reference: Any, principal: Principal, attachments: Dict[str, str], address: str):
        self.reference = reference
        self.principal = principal
        self.attachments = attachments
        self.address = address

    def invoke(self, proxy: Any, method: Any, *args, **kwargs):
        return Mesh.context_safe(lambda: self.safe_invoke(proxy, method, *args, **kwargs))

    def safe_invoke(self, proxy: Any, method: Any, *args, **kwargs):
        overrides = self.get_override_attachments()
        routable = tool.required(self.principal.inst_id) or tool.required(self.principal.node_id)
        try:
            if tool.required(self.address):
                Mesh.context().set_attribute(Mesh.REMOTE, self.address)
            if tool.required(self.attachments):
                for (k, v) in self.attachments.items():
                    Mesh.context().get_attachments().__setitem__(k, v)
            if routable:
                Mesh.context().get_principals().append(self.principal)

            executor = getattr(self.reference, method.__name__, method)
            if executor == method:
                return executor(self.reference, *args, **kwargs)
            return executor(*args, **kwargs)
        finally:
            for (k, v) in overrides.items():
                Mesh.context().get_attachments().__setitem__(k, v)
            if routable:
                Mesh.context().get_principals().pop()

    def get_override_attachments(self):
        if tool.optional(self.attachments):
            return {}

        overrides = {}
        for (k, v) in self.attachments.items():
            overrides[k] = Mesh.context().get_attachments().get(k, '')

        return overrides


@spi("mesh")
class MeshRoutable(Routable):

    def __init__(self, reference: T = None, attachments: Dict[str, str] = None, address: str = ""):
        self.ref = reference
        self.attachments = attachments if attachments is not None else {}
        self.address = address

    def within(self, key: str, value: str) -> Routable[T]:
        return self.with_map({key: value})

    def with_map(self, attachments: Dict[str, str]) -> Routable[T]:
        if not attachments or attachments.__len__() < 1:
            return self
        kvs = {}
        if self.attachments.__len__() > 0:
            for (k, v) in self.attachments.items():
                kvs[k] = v
        for (k, v) in attachments.items():
            kvs[k] = v

        return MeshRoutable(self.ref, kvs)

    def with_address(self, address: str) -> Routable[T]:
        return MeshRoutable(self.ref, self.attachments, address)

    def local(self) -> T:
        interfaces = Proxy.get_interfaces(self.ref.__class__)
        return Proxy(interfaces, MeshRouter(self.ref, Principal(), self.attachments, self.address))

    def any(self, principal: Principal) -> T:
        if tool.optional(principal.node_id) and tool.optional(principal.inst_id):
            raise ValidationException("Route key both cant be empty.")
        interfaces = Proxy.get_interfaces(self.ref.__class__)
        return Proxy(interfaces, MeshRouter(self.ref, principal, self.attachments, self.address))

    def any_inst(self, inst_id: str) -> T:
        principal = Principal()
        principal.inst_id = inst_id
        return self.any(principal)

    def many(self, principals: List[Principal]) -> List[T]:
        if tool.optional(principals):
            return []
        references = []
        for principal in principals:
            references.append(self.any(principal))
        return references

    def many_inst(self, inst_ids: List[str]) -> List[T]:
        if tool.optional(inst_ids):
            return []
        principals = []
        for inst_id in inst_ids:
            principal = Principal()
            principal.inst_id = inst_id
            principals.append(principal)
        return self.many(principals)



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

from abc import ABC, abstractmethod
from typing import Dict

from mesh.kinds.captcha import Captcha
from mesh.macro import spi, mpi


@spi("mesh")
class Graphics(ABC):

    @abstractmethod
    @mpi("mesh.graphics.captcha.apply")
    def captcha(self, kind: str, features: Dict[str, str]) -> Captcha:
        """
        Apply a graphics captcha.
        :param kind:
        :param features:
        :return:
        """
        pass

    @abstractmethod
    @mpi("mesh.graphics.captcha.verify")
    def verify(self, mno: str, value: str) -> bool:
        """
        Verify a graphics captcha value.
        :param mno:
        :param value:
        :return:
        """
        pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#
from mesh.environ.cache import Cache, CallableMember
from mesh.environ.environ import System, Addrs, URI, Mode

__all__ = ("Cache", "CallableMember", "System", "URI", "Addrs", "Mode")


def init():
    """ init function """
    pass



#
# Copyright (c) 2000, 2099, trustbe and/or its affiliates. All rights reserved.
# TRUSTBE PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
#
#

import collections
import typing

try:
    # Python >=3.8 should have these functions already
    from typing import get_args as _get_args
    from typing import get_origin as _get_origin
except ImportError:
    if hasattr(typing, '_GenericAlias'):  # Python 3.7

        def _get_origin(tp):
            """Copied from the Python 3.8 typing module"""
            if isinstance(tp, typing._GenericAlias):
                return tp.__origin__
            if tp is typing.Generic:
                return typing.Generic
            return None


        def _get_args(tp):
            """Copied from the Python 3.8 typing module"""
            if isinstance(tp, typing._GenericAlias):
                res = tp.__args__
                if (
                        get_origin(tp) is collections.abc.Callable
                        and res[0] is not Ellipsis
                ):
                    res = (list(res[:-1]), res[-1])
                return res
            return ()

    else:  # Python <3.7

        def _resolve_via_mro(tp):
            if hasattr(tp, '__mro__'):
                for t in tp.__mro__:
                    if t.__module__ in ('builtins', '__builtin__') and t is not object:
                        return t
            return tp


        def _get_origin(tp):
            """Emulate the behaviour of Python 3.8 typing module"""
            if isinstance(tp, typing._ClassVar):
                return typing.ClassVar
            elif isinstance(tp, typing._Union):
                return typing.Union
            elif isinstance(tp, typing.GenericMeta):
                if hasattr(tp, '_gorg'):
                    return _resolve_via_mro(tp._gorg)
                else:
                    while tp.__origin__ is not None:
                        tp = tp.__origin__
                    return _resolve_via_mro(tp)
            elif hasattr(typing, '_Literal') and isinstance(tp, typing._Literal):
                return typing.Literal


        def _normalize_arg(args):
            if isinstance(args, tuple) and len(args) > 1:
                base, rest = args[0], tuple(_normalize_arg(arg) for arg in args[1:])
                if isinstance(base, typing.CallableMeta):
                    return typing.Callable[[list(rest[:-1])], rest[-1]]
                elif isinstance(base, (typing.GenericMeta, typing._Union)):
                    return base[rest]
            return args


        def _get_args(tp):
            """Emulate the behaviour of Python 3.8 typing module"""
            if isinstance(tp, typing._ClassVar):
                return (tp.__type__,)
            elif hasattr(tp, '_subs_tree'):
                tree = tp._subs_tree()
                if isinstance(tree, tuple) and len(tree) > 1:
                    if isinstance(tree[0], typing.CallableMeta) and len(tree) == 2:
                        return ([], _normalize_arg(tree[1]))
                    return tuple(_normalize_arg(arg) for arg in tree[1:])
            return ()


class Compatible:

    @staticmethod
    def get_origin(tp):
        """
        Get the unsubscripted version of a type.
        This supports generic types, Callable, Tuple, Union, Literal, Final and
        ClassVar. Returns None for unsupported types.
        Examples:
            get_origin(Literal[42]) is Literal
            get_origin(int) is None
            get_origin(ClassVar[int]) is ClassVar
            get_origin(Generic) is Generic
            get_origin(Generic[T]) is Generic
            get_origin(Union[T, int]) is Union
            get_origin(List[Tuple[T, T]][int]) == list
        """
        return _get_origin(tp)

    @staticmethod
    def get_args(tp):
        """
        Get type arguments with all substitutions performed.
        For unions, basic simplifications used by Union constructor are performed.
        Examples:
            get_args(Dict[str, int]) == (str, int)
            get_args(int) == ()
            get_args(Union[int, Union[T, int], str][int]) == (int, str)
            get_args(Union[int, Tuple[T, int]][str]) == (int, Tuple[str, int])
            get_args(Callable[[], T][int]) == ([], int)
        """
        return _get_args(tp)
